/*	$NetBSD$	*/
/*	$NetBSD/i386: isa.c,v 1.107 2000/03/29 03:43:31 simonb Exp $	*/

/*-
 * Copyright (c) 1998 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Charles M. Hannum.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *        This product includes software developed by the NetBSD
 *        Foundation, Inc. and its contributors.
 * 4. Neither the name of The NetBSD Foundation nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/device.h>
#include <sys/syslog.h>
#include <sys/proc.h>
#include <sys/mbuf.h>

#include <machine/cpufunc.h>
#include <machine/i8259.h>
#include <machine/intr.h>
#include <machine/pio.h>

#include <uvm/uvm.h>

#include <pc98/cbus/cbusreg.h>
#include <pc98/cbus/cbusvar.h>

#include "cbusdma.h"

#if 0
#include "cbuspnp.h"
#if NCBUSPNP > 0
#include <pc98/cbus/cbuspnpreg.h>
#include <pc98/cbus/cbuspnpvar.h>
#endif /* NCBUSPNP > 0 */
#endif /* 0 */

static int cbusmatch(struct device *, struct cfdata *, void *);
static void cbusattach(struct device *, struct device *, void *);
static int cbusprint(void *, const char *);

CFATTACH_DECL(cbus, sizeof(struct cbus_softc),
    cbusmatch, cbusattach, NULL, NULL);

static int cbussearch(struct device *, struct cfdata *, void *);

static int
cbusmatch(struct device *parent, struct cfdata *cf, void *aux)
{
	struct neccbus_attach_args *cba = aux;

	if (strcmp(cba->cba_busname, cf->cf_name))
		return (0);

	/* XXX check other indicators */

        return (1);
}

static void
cbusattach(struct device *parent, struct device *self, void *aux)
{
	struct cbus_softc *sc = (struct cbus_softc *)self;
	struct neccbus_attach_args *cba = aux;

	cbus_attach_hook(parent, self, cba);
	printf("\n");

	sc->sc_iot = cba->cba_iot;
	sc->sc_memt = cba->cba_memt;
	sc->sc_dmat = cba->cba_dmat;
	sc->sc_ic = cba->cba_ic;

#if NCBUSPNP > 0
	/*
	 * Reset PNP cards that the bios configured for us
	 */
	cbuspnp_cbus_attach_hook(sc);
#endif

#if NCBUSDMA > 0
	/*
	 * Initialize our DMA state.
	 */
	cbus_dmainit(sc->sc_ic, sc->sc_iot, sc->sc_dmat, self);
#endif

	config_search(cbussearch, self, NULL);
}

static int
cbusprint(void *aux, const char *cbus)
{
	struct cbus_attach_args *ca = aux;

	if (ca->ca_iosize)
		printf(" port 0x%x", ca->ca_iobase);
	if (ca->ca_iosize > 1)
		printf("-0x%x", ca->ca_iobase + ca->ca_iosize - 1);
	if (ca->ca_msize)
		printf(" iomem 0x%x", ca->ca_maddr);
	if (ca->ca_msize > 1)
		printf("-0x%x", ca->ca_maddr + ca->ca_msize - 1);
	if (ca->ca_irq != IRQUNK)
		printf(" irq %d", ca->ca_irq);
	if (ca->ca_drq != DRQUNK)
		printf(" drq %d", ca->ca_drq);
	if (ca->ca_drq2 != DRQUNK)
		printf(" drq2 %d", ca->ca_drq2);
	return (UNCONF);
}

static int
cbussearch(struct device *parent, struct cfdata *cf, void *aux)
{
	struct cbus_softc *sc = (struct cbus_softc *)parent;
	struct cbus_attach_args ca;
	int tryagain;

	do {
		ca.ca_iot = sc->sc_iot;
		ca.ca_memt = sc->sc_memt;
		ca.ca_dmat = sc->sc_dmat;
		ca.ca_ic = sc->sc_ic;
		ca.ca_iobase = cf->cf_iobase;
		ca.ca_iosize = 0x666; /* cf->cf_iosize; */
		ca.ca_maddr = cf->cf_maddr;
		ca.ca_msize = cf->cf_msize;
		ca.ca_irq = cf->cf_irq == 2 ? 9 : cf->cf_irq;
		ca.ca_drq = cf->cf_drq;
		ca.ca_drq2 = cf->cf_drq2;

		tryagain = 0;
		if (config_match(parent, cf, &ca) > 0) {
			config_attach(parent, cf, &ca, cbusprint);
			tryagain = (cf->cf_fstate == FSTATE_STAR);
		}
	} while (tryagain);

	return (0);
}

char *
cbus_intr_typename(int type)
{

	switch (type) {
        case IST_NONE :
		return ("none");
        case IST_PULSE:
		return ("pulsed");
        case IST_EDGE:
		return ("edge-triggered");
        case IST_LEVEL:
		return ("level-triggered");
	default:
		panic("cbus_intr_typename: invalid type %d", type);
	}
}

#define	LEGAL_IRQ(x)	((x) >= 0 && (x) < ICU_LEN && (x) != 7)

int
cbus_intr_alloc(cbus_chipset_tag_t cc, int mask, int type, int *irq)
{
	int i, tmp, bestirq, count;
	struct intrhand **p, *q;
	struct intrsource *isp;
	struct cpu_info *ci;

	if (type == IST_NONE)
		panic("intr_alloc: bogus type");

	ci = &cpu_info_primary;

	bestirq = -1;
	count = -1;

	/* some interrupts should never be dynamically allocated */
	mask &= 0xdef8;

	/*
	 * XXX some interrupts will be used later (6 for fdc, 12 for pms).
	 * the right answer is to do "breadth-first" searching of devices.
	 */
	mask &= 0xefbf;

	simple_lock(&ci->ci_slock);

	for (i = 0; i < NUM_LEGACY_IRQS; i++) {
		if (LEGAL_IRQ(i) == 0 || (mask & (1<<i)) == 0)
			continue;

		isp = ci->ci_isources[i];
		if (isp == NULL) {
			/*
			 * if nothing's using the irq, just return it
			 */
			*irq = i;
			simple_unlock(&ci->ci_slock);
			return (0);
		}

		switch(isp->is_type) {
		case IST_EDGE:
		case IST_LEVEL:
			if (type != isp->is_type)
				continue;
			/*
			 * if the irq is shareable, count the number of other
			 * handlers, and if it's smaller than the last irq like
			 * this, remember it
			 *
			 * XXX We should probably also consider the
			 * interrupt level and stick IPL_TTY with other
			 * IPL_TTY, etc.
			 */
			for (p = &isp->is_handlers, tmp = 0; (q = *p) != NULL;
			     p = &q->ih_next, tmp++)
				continue;
			if ((bestirq == -1) || (count > tmp)) {
				bestirq = i;
				count = tmp;
			}
			break;

		case IST_PULSE:
			/* this just isn't shareable */
			continue;
		}
	}

	simple_unlock(&ci->ci_slock);

	if (bestirq == -1)
		return (1);

	*irq = bestirq;

	return (0);
}

/*
 * Set up an interrupt handler to start being called.
 * XXX PRONE TO RACE CONDITIONS, UGLY, 'INTERESTING' INSERTION ALGORITHM.
 */
void *
cbus_intr_establish(cbus_chipset_tag_t cc, int irq, int type, int level,
	int (*ih_fun)(void *), void *ih_arg)
{
	struct pic *pic;
	int pin;

	pin = irq;
	pic = &i8259_pic;

	return intr_establish(irq, pic, pin, type, level, ih_fun, ih_arg);
}

/*
 * Deregister an interrupt handler.
 */
void
cbus_intr_disestablish(cbus_chipset_tag_t cc, void *arg)
{
	struct intrhand *ih = arg;

	if (!LEGAL_IRQ(ih->ih_pin))
		panic("intr_disestablish: bogus irq");

	intr_disestablish(ih);
}

void
cbus_attach_hook(struct device *parent, struct device *self,
    struct neccbus_attach_args *cba)
{
	static struct pc98_cbus_chipset pc98_cbus_chipset;
	extern int cbus_has_been_seen;

	/*
	 * Notify others that might need to know that the ISA bus
	 * has now been attached.
	 */
	if (cbus_has_been_seen)
		panic("cbus_attach_hook: CBus already seen!");
	cbus_has_been_seen = 1;

	/*
	 * Since we can only have one ISA bus, we just use a single
	 * statically allocated ISA chipset structure.  Pass it up
	 * now.
	 */
	cba->cba_ic = &pc98_cbus_chipset;
}

int
cbus_mem_alloc(bus_space_tag_t t, bus_size_t size, bus_size_t align,
    bus_addr_t boundary, int flags, bus_addr_t *addrp, bus_space_handle_t *bshp)
{

	/*
	 * Allocate physical address space in the ISA hole.
	 */
	return (bus_space_alloc(t, IOM_BEGIN, IOM_END - 1, size, align,
	    boundary, flags, addrp, bshp));
}

void
cbus_mem_free(bus_space_tag_t t, bus_space_handle_t bsh, bus_size_t size)
{

	bus_space_free(t, bsh, size);
}
