/*	$NetBSD$	*/

/*
 * Copyright (c) 1997 Marc Horowitz.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Marc Horowitz.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__KERNEL_RCSID(0, "$NetBSD$");

#define	PCICCBUSDEBUG

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/device.h>
#include <sys/extent.h>
#include <sys/malloc.h>

#include <machine/bus.h>
#include <machine/intr.h>

#include <dev/pcmcia/pcmciareg.h>
#include <dev/pcmcia/pcmciavar.h>
#include <dev/pcmcia/pcmciachip.h>

#include <dev/ic/i82365reg.h>
#include <dev/ic/i82365var.h>

#include <pc98/cbus/cbusreg.h>
#include <pc98/cbus/cbusvar.h>
#include <pc98/cbus/i82365_cbusvar.h>

#ifdef PCICCBUSDEBUG
int	pciccbus_debug = 0;
#define	DPRINTF(arg) if (pciccbus_debug) printf arg;
#else
#define	DPRINTF(arg)
#endif

#undef	PCIC_IOSIZE
#define	PCIC_IOSIZE	4

int	pcic_cbus_probe __P((struct device *, struct cfdata *, void *));
void	pcic_cbus_attach __P((struct device *, struct device *, void *));

CFATTACH_DECL(pcic_cbus, sizeof(struct pcic_cbus_softc),
    pcic_cbus_probe, pcic_cbus_attach, NULL, NULL);

static struct pcmcia_chip_functions pcic_cbus_functions = {
	pcic_chip_mem_alloc,
	pcic_chip_mem_free,
	pcic_chip_mem_map,
	pcic_chip_mem_unmap,

	pcic_chip_io_alloc,
	pcic_chip_io_free,
	pcic_chip_io_map,
	pcic_chip_io_unmap,

	pcic_cbus_chip_intr_establish,
	pcic_cbus_chip_intr_disestablish,

	pcic_chip_socket_enable,
	pcic_chip_socket_disable,
};

int
pcic_cbus_probe(parent, match, aux)
	struct device *parent;
	struct cfdata *match;
	void *aux;
{
	struct cbus_attach_args *ca = aux;
	bus_space_tag_t iot = ca->ca_iot;
	bus_space_handle_t ioh, memh;
	int val, found, msize;

#if 0
	if (ca->ca_nio < 1)
		return (0);
	if (ca->ca_niomem < 1)
		return (0);

	if (CBUS_DIRECT_CONFIG(ca))
		return (0);
#endif

	/* Disallow wildcarded i/o address. */
	if (ca->ca_ioaddr == CBUSCF_PORT_DEFAULT)
		return (0);
	if (ca->ca_maddr == CBUSCF_IOMEM_DEFAULT)
		return (0);

	if (bus_space_map(iot, ca->ca_ioaddr, PCIC_IOSIZE, 0, &ioh))
		return (0);

	if (ca->ca_msize == CBUSCF_IOSIZ_DEFAULT)
		msize = PCIC_MEMSIZE;
	else
		msize = ca->ca_msize;

	if (bus_space_map(ca->ca_memt, ca->ca_maddr, msize, 0, &memh)) {
		bus_space_unmap(iot, ioh, PCIC_IOSIZE);
		return (0);
	}

	found = 0;

	/*
	 * this could be done with a loop, but it would violate the
	 * abstraction
	 */
	bus_space_write_1(iot, ioh, PCIC_REG_INDEX, C0SA + PCIC_IDENT);
	val = bus_space_read_1(iot, ioh, PCIC_REG_DATA);
	if (pcic_ident_ok(val))
		found++;

	bus_space_write_1(iot, ioh, PCIC_REG_INDEX, C0SB + PCIC_IDENT);
	val = bus_space_read_1(iot, ioh, PCIC_REG_DATA);
	if (pcic_ident_ok(val))
		found++;

	bus_space_write_1(iot, ioh, PCIC_REG_INDEX, C1SA + PCIC_IDENT);
	val = bus_space_read_1(iot, ioh, PCIC_REG_DATA);
	if (pcic_ident_ok(val))
		found++;

	bus_space_write_1(iot, ioh, PCIC_REG_INDEX, C1SB + PCIC_IDENT);
	val = bus_space_read_1(iot, ioh, PCIC_REG_DATA);
	if (pcic_ident_ok(val))
		found++;

	bus_space_unmap(iot, ioh, PCIC_IOSIZE);
	bus_space_unmap(ca->ca_memt, memh, msize);

	if (!found)
		return (0);

#if 1
	ca->ca_iosize = PCIC_IOSIZE;
	ca->ca_msize = msize;
#else
	ca->ca_nio = 1;
	ca->ca_iosize = PCIC_IOSIZE;

	ca->ca_niomem = 1;
	ca->ca_msize = msize;

	/* IRQ is special. */
	ca->ca_ndrq = 0;
#endif

	return (1);
}

void
pcic_cbus_attach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct pcic_softc *sc = (void *) self;
	struct pcic_cbus_softc *csc = (void *) self;
	struct cbus_attach_args *ca = aux;
	cbus_chipset_tag_t cc = ca->ca_ic;
	bus_space_tag_t iot = ca->ca_iot;
	bus_space_tag_t memt = ca->ca_memt;
	bus_space_handle_t ioh;
	bus_space_handle_t memh;

	/* Map i/o space. */
	if (bus_space_map(iot, ca->ca_ioaddr, PCIC_IOSIZE, 0, &ioh)) {
		printf(": can't map i/o space\n");
		return;
	}

	/* Map mem space. */
	if (bus_space_map(memt, ca->ca_maddr, ca->ca_msize, 0, &memh)) {
		printf(": can't map mem space\n");
		return;
	}

	sc->membase = ca->ca_maddr;
	sc->subregionmask = (1 << (ca->ca_msize / PCIC_MEM_PAGESIZE)) - 1;

	csc->sc_ic = cc;
	sc->pct = (pcmcia_chipset_tag_t) & pcic_cbus_functions;

	sc->iot = iot;
	sc->ioh = ioh;
	sc->memt = memt;
	sc->memh = memh;
#if 0
	if (ca->ca_nirq > 0)
		sc->irq = ca->ca_irq;
	else
#endif
		sc->irq = CBUSCF_IRQ_DEFAULT;

	printf("\n");

	pcic_attach(sc);
	pcic_cbus_bus_width_probe(sc, iot, ioh, ca->ca_ioaddr, PCIC_IOSIZE);
	pcic_attach_sockets(sc);

	config_interrupts(self, pcic_cbus_config_interrupts);
}
