/*	$NetBSD$	*/
/*	$NetBSD/i386: pcib.c,v 1.25 1999/09/20 16:07:45 drochner Exp $	*/

/*-
 * Copyright (c) 1996, 1998 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Jason R. Thorpe.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *        This product includes software developed by the NetBSD
 *        Foundation, Inc. and its contributors.
 * 4. Neither the name of The NetBSD Foundation nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/device.h>

#include <machine/bus.h>

#include <pc98/cbus/cbusvar.h>

#include <dev/pci/pcivar.h>
#include <dev/pci/pcireg.h>

#include <dev/pci/pcidevs.h>

#include "cbus.h"

int	pccbmatch __P((struct device *, struct cfdata *, void *));
void	pccbattach __P((struct device *, struct device *, void *));

CFATTACH_DECL(pccb, sizeof(struct device),
	pccbmatch, pccbattach, NULL, NULL);

void	pccb_callback __P((struct device *));
int	pccb_print __P((void *, const char *));

int
pccbmatch(parent, match, aux)
	struct device *parent;
	struct cfdata *match;
	void *aux;
{
	struct pci_attach_args *pa = aux;

	switch (PCI_VENDOR(pa->pa_id)) {
	case PCI_VENDOR_NEC:
		switch (PCI_PRODUCT(pa->pa_id)) {
		case PCI_PRODUCT_NEC_CBUSBRIDGE1:
		case PCI_PRODUCT_NEC_CBUSBRIDGE2:
			return (1);
		}
		break;
	}

	return (0);
}

void
pccbattach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct pci_attach_args *pa = aux;
	char devinfo[256];

	printf("\n");

	/*
	 * Just print out a description and defer configuration
	 * until all PCI devices have been attached.
	 */
	pci_devinfo(pa->pa_id, pa->pa_class, 0, devinfo);
	printf("%s: %s (rev. 0x%02x)\n", self->dv_xname, devinfo,
	    PCI_REVISION(pa->pa_class));

	config_defer(self, pccb_callback);
}

void
pccb_callback(self)
	struct device *self;
{
	struct neccbus_attach_args cba;

	/*
	 * Attach the ISA bus behind this bridge.
	 */
	memset(&cba, 0, sizeof(cba));
	cba.cba_busname = "cbus";
	cba.cba_iot = PC98_BUS_SPACE_IO;
	cba.cba_memt = PC98_BUS_SPACE_MEM;
#if NCBUSDMA > 0
	cba.cba_dmat = &cbus_bus_dma_tag;
#endif
	config_found(self, &cba, pccb_print);
}

int
pccb_print(aux, pnp)
	void *aux;
	const char *pnp;
{

	/* Only ISAs can attach to pccb's; easy. */
	if (pnp)
		printf("cbus at %s", pnp);
	return (UNCONF);
}
